/* mincregress.c                                                              */
/*                                                                            */
/* do various regressions on MINC files                                       */
/*                                                                            */
/* Andrew Janke - a.janke@gmail.com                                           */
/* Mark Griffin - mgriffin@bic.mni.mcgill.ca                                  */
/*                                                                            */
/* Copyright Andrew Janke and Mark Griffin, McConnell Brain Imaging Centre    */
/* Permission to use, copy, modify, and distribute this software and its      */
/* documentation for any purpose and without fee is hereby granted,           */
/* provided that the above copyright notice appear in all copies.  The        */
/* author and McGill University make no representations about the             */
/* suitability of this software for any purpose.  It is provided "as is"      */
/* without express or implied warranty.                                       */

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <ParseArgv.h>
#include <time_stamp.h>
#include <voxel_loop.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <minc_vector_io.h>
#include <volume_io.h>

#ifndef  FALSE
#define  FALSE  0
#endif

#ifndef  TRUE
#define  TRUE   1
#endif

#define MAX_CONTRASTS 100

/* typedefs */
typedef struct {
   /* input parameters */
   int      masking;
   double   mask_val;
   int      mask_idx;

   gsl_matrix *svd_u, *svd_v;
   gsl_vector *svd_s, *svd_x, *svd_b;

   } Loop_Data;

/* function prototypes */
void     do_loop(void *caller_data, long num_voxels, int input_num_buffers,
                 int input_vector_length, double *input_data[],
                 int output_num_buffers, int output_vector_length,
                 double *output_data[], Loop_Info * loop_info);
void     print_version_info(void);

/* Argument variables and table */
static int verbose = FALSE;
static int clobber = FALSE;
static int max_buffer = 4 * 1024;
static char *mask_fname = NULL;
static char *contrast_fname_string = NULL;
static Loop_Data md = {
   FALSE, 1.0, 0,
   NULL, NULL,
   NULL, NULL, NULL
   };

static ArgvInfo argTable[] = {
   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL,
    "General options:"},
   {"-version", ARGV_FUNC, (char *)print_version_info, (char *)NULL,
    "print version info and exit."},
   {"-verbose", ARGV_CONSTANT, (char *)TRUE, (char *)&verbose,
    "print out extra information."},
   {"-clobber", ARGV_CONSTANT, (char *)TRUE, (char *)&clobber,
    "clobber existing files."},
   {"-max_buffer", ARGV_INT, (char *)1, (char *)&max_buffer,
    "maximum size of buffers (in kbytes)"},
   {"-mask", ARGV_STRING, (char *)1, (char *)&mask_fname,
    "select voxels within the specified mask"},
   {"-mask_val", ARGV_FLOAT, (char *)1, (char *)&(md.mask_val),
    "mask value to use"},

   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL,
    "\nContrasts:"},
   {"-contrasts", ARGV_STRING, (char *)1, (char *)&contrast_fname_string,
    "comma separated list of contrast filenames"},

   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL,
    "\nOutput Options:"},

   {NULL, ARGV_HELP, NULL, NULL, ""},
   {NULL, ARGV_END, NULL, NULL, NULL}
   };

int main(int argc, char *argv[])
{
   char   **infiles;
   char   **outfiles;
   char    *out_base;
   int      n_infiles, n_outfiles;
   char    *arg_string;
   Loop_Options *loop_opts;
   int      i, j;
   char    *ptr;

   int      n_frames;

   int      n_contrasts;
   char    *contrast_fnames[MAX_CONTRASTS];

   MINC_Vector *tmp;

   gsl_matrix *contrasts;

   /* Save time stamp and args */
   arg_string = time_stamp(argc, argv);

   /* get arguments */
   if(ParseArgv(&argc, argv, argTable, 0) || (argc < 2)){
      fprintf(stderr, "\nUsage: %s [options] <frame1.mnc> <frame2.mnc> ... <out_base>\n",
              argv[0]);
      fprintf(stderr, "       %s -help\n\n", argv[0]);
      exit(EXIT_FAILURE);
      }

   /* get infile names */
   n_infiles = argc - 2;
   infiles = (char **)malloc(sizeof(char *) * (n_infiles + 1));   /* + 1 for mask */
   for(i = 0; i < n_infiles; i++){
      infiles[i] = argv[i + 1];
      }
   n_frames = n_infiles;

   /* get output base */
   out_base = argv[argc-1];

   /* set up mask file */
   if(mask_fname != NULL){
      infiles[n_infiles] = mask_fname;
      md.masking = TRUE;
      md.mask_idx = n_infiles;
      n_infiles++;
      }

   /* check for the infile(s) */
   for(i = 0; i < n_infiles; i++){
      if(access(infiles[i], F_OK) != 0){
         fprintf(stderr, "%s: Couldn't find input file: %s\n\n", argv[0], infiles[i]);
         exit(EXIT_FAILURE);
         }
      }

   /* read in input contrasts (regressor vectors) */
#define ARG_SEPARATOR ','
   ptr = contrast_fname_string;
   n_contrasts = 0;

   /* Loop through string looking for space or comma-separated names */
   while((n_contrasts < MAX_CONTRASTS) && (*ptr != '\0')){

      /* Get string */
      contrast_fnames[n_contrasts] = ptr;

      /* Search for end of dimension name */
      while((*ptr != ARG_SEPARATOR) && (*ptr != '\0')){
         ptr++;
         }

      if(*ptr != '\0'){
         *ptr = '\0';
         ptr++;
         }
      n_contrasts++;
      }

   /* check that n_frames >= n_contrasts */
   if(n_frames < n_contrasts){
      fprintf(stderr, "%s: Must not have less contrasts (%d) than frames (%d)\n\n",
              argv[0], n_contrasts, n_frames);
      exit(EXIT_FAILURE);
      }

   /* allocate space for our contrasts matrix */
   contrasts = gsl_matrix_alloc(n_frames, n_contrasts);

   /* read in the minc vectors and convert to gsl_vectors */
   tmp = new_MINC_Vector(0);

   for(j = 0; j < n_contrasts; j++){

      if(input_MINC_Vector(contrast_fnames[j], tmp) == OK){

         /* check that this one is the right length */
         if(tmp->size != n_frames){
            fprintf(stderr,
                    "%s: size of input vector %s (%d) must equal the number of frames (%d) \n\n",
                    argv[0], contrast_fnames[j], tmp->size, n_frames);
            exit(EXIT_FAILURE);
            }

         for(i = 0; i < tmp->size; i++){
            gsl_matrix_set(contrasts, i, j, tmp->V[i]);
            }
         }
      else {
         fprintf(stderr, "%s: Couldn't read in: %s\n\n", argv[0], contrast_fnames[i]);
         exit(EXIT_FAILURE);
         }
      }

   free_MINC_Vector(tmp);

   /* set up and check for output files */
   n_outfiles = n_contrasts;
   outfiles = (char **)malloc(sizeof(char *) * n_outfiles);
   for(i = 0; i < n_outfiles; i++){
      outfiles[i] = (char *)malloc((strlen(out_base) + strlen(contrast_fnames[i]) + 5)
                                   * sizeof(char));
      sprintf(outfiles[i], "%s.%s.mnc", out_base, contrast_fnames[i]);

      if(access(outfiles[i], F_OK) == 0 && !clobber){
         fprintf(stderr, "%s: %s exists, use -clobber to overwrite\n\n",
                 argv[0], outfiles[i]);
         exit(EXIT_FAILURE);
         }
      }

   /* set up math data */
   md.svd_u = contrasts;
   md.svd_v = gsl_matrix_alloc(n_contrasts, n_contrasts);
   md.svd_s = gsl_vector_alloc(n_contrasts);
   md.svd_x = gsl_vector_alloc(n_contrasts);
   md.svd_b = gsl_vector_alloc(n_frames);

   /* do the SVD decomposition */
   gsl_linalg_SV_decomp(md.svd_u, md.svd_v, md.svd_s, md.svd_x);
   
   /* set up voxel loop options */
   loop_opts = create_loop_options();
   set_loop_verbose(loop_opts, FALSE);
   set_loop_clobber(loop_opts, clobber);
   set_loop_buffer_size(loop_opts, (long)1024 * max_buffer);

   /* do the sampling */
   voxel_loop(n_infiles, infiles, n_outfiles, outfiles, arg_string,
              loop_opts, do_loop, (void *)&md);

   /* tidy up */
   free_loop_options(loop_opts);

   // lots of gsl_free

   return (EXIT_SUCCESS);
   }

/* get points from file(s), write out to an input FP */
void do_loop(void *caller_data, long num_voxels, int input_num_buffers,
             int input_vector_length, double *input_data[], int output_num_buffers,
             int output_vector_length, double *output_data[], Loop_Info * loop_info)
{
   Loop_Data *md = (Loop_Data *) caller_data;
   int      i, idim, ivox;
   int      n_infiles;
   int      do_sample;
   double   mask_value;
   int      dim_index;
   long     index[MAX_VAR_DIMS];

   /* shut the compiler up */
   // (void)output_num_buffers;
   // (void)output_vector_length;

   n_infiles = (md->masking) ? input_num_buffers - 1 : input_num_buffers;

   /* for each voxel */
   for(ivox = 0; ivox < num_voxels * input_vector_length; ivox++){
      /* nasty way that works for masking or not */
      mask_value = 0;
      if(!md->masking ||
         (md->masking && fabs(input_data[md->mask_idx][ivox] - md->mask_val) < 0.5)){

         /* set up svd_b */
         for(i = 0; i < n_infiles; i++){
            gsl_vector_set(md->svd_b, i, input_data[i][ivox]);
            }



         /* solve ax=b for x */
         gsl_linalg_SV_solve(md->svd_u, md->svd_v, md->svd_s, md->svd_b, md->svd_x);

         /* write results back out */
         for(i = 0; i < output_num_buffers; i++){
            output_data[i][ivox] = gsl_vector_get(md->svd_x, i);
            }

         /* get measure of fittness from mincdti */

         }
      }

   }

void print_version_info(void)
{
   fprintf(stdout, "\n");
   fprintf(stdout, "%s version %s\n", PACKAGE, VERSION);
   fprintf(stdout, "Comments to %s\n", PACKAGE_BUGREPORT);
   fprintf(stdout, "\n");
   exit(EXIT_SUCCESS);
   }
